﻿//-----------------------------------------------------------------------
using System;
using System.IO;
using System.IO.Ports;
using System.Diagnostics;
using System.Threading;
using System.Collections.Generic;

namespace Serialcode.bob
{

    public class Serialcode : IDisposable
    {
        private const int READ_BUFFER_SIZE = 128;
        private const int MAX_DATA_LENGTH = 16;
        private const int MAX_RESPONSE_TIME = 250;
        private const int LINE_LENGTH = 16;
        private const int MAX_PACKET_LENGTH = MAX_DATA_LENGTH + 4;
        private const byte NORMAL_RESPONSE = 0x40;
        private const byte NORMAL_REPORT = 0x80;
        private const byte ERROR_RESPONSE = 0xC0;
        private const byte KEY_ACTIVITY_REPORT = 0x80;
        private const byte FAN_SPEED_REPORT = 0x81;
        private const byte TEMPATURE_SENSOR_REPORT = 0x82;
        ushort CRC_SEED = 0xFFFF;
        private byte[] packetRcvBuffer;
        private byte[] packetXMitBuffer;
        private bool _timedOut;
        private Thread receiveThread;
        private Thread eventThread;
        //private LCDPacket responsePacket;
        //private Queue<LCDPacket> reportQueue;
        //private Object responseSignal;
        //private Object reportSignal;
  
        // comPort
        private SerialPort com;

        public void SerialLCD()
        {
            // comPort
            com = new SerialPort("COM3", 9600);
            com.ReadTimeout = 3000;
            com.WriteTimeout = 3000;
            com.Open();

            packetRcvBuffer = new byte[MAX_PACKET_LENGTH];
            packetXMitBuffer = new byte[MAX_PACKET_LENGTH];
            //responseSignal = new Object();
            //reportSignal = new Object();
            eventThread = new Thread(new ThreadStart(ReportEventHandler));
            eventThread.Start();
            receiveThread = new Thread(new ThreadStart(Receive));
            receiveThread.Start();
        }

        public bool TimedOut { get { return _timedOut; } set { _timedOut = value; } }
        public byte[] Ping(byte[] data)
        {
            byte type = 0;

            if (16 < data.Length)
            {
                throw new ArgumentException("data", "must have 16 items or less");
            }

            return SendReturnData(type, (byte)data.Length, data);
        }
        public bool IsOpen
        {
            get
            {
                if (com == null)
                {
                    return false;
                }
                else
                {
                    return com.IsOpen;
                }
            }
        }

//---------------------------------------------------------------
/*        public byte[] GetHardwareFirmwareVersion()
        {
            byte type = 1;

            return SendReturnData(type, 0, null);
        }
        public bool WriteUserFlashArea(byte[] data)
        {
            byte type = 2;

            data = CreateByteArrayOfLength(data, 16);
            return SendReturnBool(type, 16, data);
        }
        public byte[] ReadUserFlashArea()
        {
            byte type = 3;

            return SendReturnData(type, 0, null);
        }
        public bool SaveState()
        {
            byte type = 4;

            return SendReturnBool(type, 0, null);
        }
        public bool Clear()
        {
            byte type = 6;

            return SendReturnBool(type, 0, null);
        } 
        public bool WriteLine1(string line)
        {
            byte type = 7;

            return WriteLine(type, line);
        }
        public bool WriteLine1(byte[] data)
        {
            byte type = 7;

            data = CreateByteArrayOfLength(data, 16);
            return SendReturnBool(type, 16, data);
        }
        public bool WriteLine2(string line)
        {
            byte type = 8;
            return WriteLine(type, line);
        }
        public bool WriteLine2(byte[] data)
        {
            byte type = 8;

            data = CreateByteArrayOfLength(data, 16);
            return SendReturnBool(type, 16, data);
        }
        private bool WriteLine(byte type, string line)
        {
            line = CreateStringOfLength(line, LINE_LENGTH);
            return SendReturnBool(type, 16, System.Text.Encoding.ASCII.GetBytes(line.ToCharArray()));
        }

        */
 
        private string CreateStringOfLength(string s, int length)
        {
            if (length < s.Length)
                s = s.Substring(length);
            else if (length > s.Length)
            {
                s = s + (new String(' ', length - s.Length));
            }

            return s;
        }
        private byte[] CreateByteArrayOfLength(byte[] bArray, int length)
        {
            if (length != bArray.Length)
            {
                int dataLength = bArray.Length > length ? length : bArray.Length;
                byte[] newbArray = new byte[length];

                Array.Copy(bArray, newbArray, dataLength);
                bArray = newbArray;
            }

            return bArray;
        }
        private byte[] SendReturnData(byte type, byte dataLength, byte[] data)
        {
            Byte packet = Send(type, dataLength, data);

            if (null != packet)
            {
                return packet.Data;
            }
            else
            {
                return null;
            }
        }

        private void Receive()
        {
            try
            {
                byte[] receiveBuffer = new byte[128];
                int bytesRead = 0;
                int bufferIndex = 0;
                int startPacketIndex = 0;
                int expectedPacketLength = -1;
                bool expectedPacketLengthIsSet = false;
                int numBytesToRead = receiveBuffer.Length;

                while (true)
                {
                    if (expectedPacketLengthIsSet || 1 >= bytesRead)
                    {
                        //If the expectedPacketLength has been or no bytes have been read
                        //This covers the case that more then 1 entire packet has been read in at a time
                        // comPort
                        try
                        {
                            bytesRead += com.Read(receiveBuffer, bufferIndex, numBytesToRead);
                            bufferIndex = startPacketIndex + bytesRead;
                        }
                        catch (TimeoutException)
                        {
                            _timedOut = true;
                        }
                    }

                    if (1 < bytesRead)
                    {
                        //The buffer has the dataLength for the packet
                        if (!expectedPacketLengthIsSet)
                        {
                            //If the expectedPacketLength has not been set for this packet
                            expectedPacketLength = receiveBuffer[(1 + startPacketIndex) % receiveBuffer.Length] + 4;
                            expectedPacketLengthIsSet = true;
                        }

                        if (bytesRead >= expectedPacketLength)
                        {
                            //The buffer has atleast as many bytes for this packet
                            AddPacket(receiveBuffer, startPacketIndex);
                            expectedPacketLengthIsSet = false;
                            if (bytesRead == expectedPacketLength)
                            {
                                //The buffer contains only the bytes for this packet
                                bytesRead = 0;
                                bufferIndex = startPacketIndex;
                            }
                            else
                            {
                                //The buffer also has bytes for the next packet
                                startPacketIndex += expectedPacketLength;
                                startPacketIndex %= receiveBuffer.Length;
                                bytesRead -= expectedPacketLength;
                                bufferIndex = startPacketIndex + bytesRead;
                            }
                        }
                    }

                    bufferIndex %= receiveBuffer.Length;
                    numBytesToRead = bufferIndex < startPacketIndex ? startPacketIndex - bufferIndex : receiveBuffer.Length - bufferIndex;
                }
            }
            catch (IOException)
            {
                // abort the thread
                System.Threading.Thread.CurrentThread.Abort();
            }
            catch (ObjectDisposedException)
            {
                if (receiveThread != null)
                {
                    receiveThread = null;
                }
            }
        }
        private void ReportEventHandler()
        {
            try
            {
                LCDPacket packet = null;

                while (true)
                {
                    while (null == packet)
                    {
                        lock (reportSignal)
                        {
                            if (0 != reportQueue.Count)
                                packet = reportQueue.Dequeue();
                            else
                                Monitor.Wait(reportSignal);
                        }
                    }

                    switch (packet.Type)
                    {
                        case KEY_ACTIVITY_REPORT:
                            ReportKeyActivityEventHandler(packet);
                            break;

                        case TEMPATURE_SENSOR_REPORT:
                            break;
                    }
                    packet = null;
                }
            }
            catch (IOException)
            {
                // abort the thread
                System.Threading.Thread.CurrentThread.Abort();
            }
            catch (ObjectDisposedException)
            {
                if (eventThread != null)
                {
                    eventThread = null;
                }
            }
        }

        private Serialcode CreatePacket(byte[] buffer, int startIndex)
        {
            byte type = buffer[startIndex];
            byte dataLength = buffer[(startIndex + 1) % buffer.Length];
            byte[] data = new byte[dataLength];
            ushort crc = 0;

            for (int i = 0; i < dataLength; i++)
            {
                data[i] = buffer[(startIndex + 2 + i) % buffer.Length];
            }

            crc |= (ushort)buffer[(startIndex + 2 + dataLength) % buffer.Length];
            crc |= (ushort)(buffer[(startIndex + 2 + dataLength + 1) % buffer.Length] << 8);
            return new LCDPacket(type, dataLength, data, crc);
        }
        private bool AddPacket(byte[] buffer, int startIndex)
        {
            LCDPacket packet = CreatePacket(buffer, startIndex);
            ushort calculatedCRC = CRCGenerator.GenerateCRC(buffer, startIndex, packet.DataLength + 2, CRC_SEED);

            switch (packet.PacketType)
            {
                case LCDPacket.LCDPacketType.NORMAL_RESPONSE:
                    AddResponsePacket(packet);
                    break;

                case LCDPacket.LCDPacketType.NORMAL_REPORT:
                    AddReportPacket(packet);
                    break;

                case LCDPacket.LCDPacketType.ERROR_RESPONSE:
                    AddResponsePacket(packet);
                    break;
            }
            if (calculatedCRC != packet.CRC)
            {
                Console.WriteLine("CRC ERROR!!!: Calculated CRC={0} Actual CRC={1}", Convert.ToString(calculatedCRC, 16), packet.CRC);
                return false;
            }

            return true;
        }

        private void AddResponsePacket(Serialcode packet)
        {
            lock (responseSignal)
            {
                responsePacket = packet;
                Monitor.Pulse(responseSignal);
            }
        }
        private void AddReportPacket(Serialcode packet)
        {
            lock (reportSignal)
            {
                reportQueue.Enqueue(packet);
                Monitor.Pulse(reportSignal);
            }
        }

       

        private void CloseThreads()
        {
            if (receiveThread != null)
            {
                receiveThread.Abort();
            }
            if (eventThread != null)
            {
                eventThread.Abort();
            }
        }

        // Public IDisposable.Dispose implementation - calls the internal helper,
        public void Dispose()
        {
            Dispose(true);
        }

        // This is "friendly" wrapper method for IDisposable.Dispose
        public void Close()
        {
            // Call the internal diposal helper
            Dispose(true);
        }

        /*
         * This class wraps a SerialPort instance that is not exposed 
         * directly to the consuming application.  The SerialPort implements 
         * IDisposable, and calling Dispose there causes the SafeHandle
         * given by the OS to the serial port to be released. 
         * 
         * 'disposing' is true if Close or Dispose have been called explicitly,
         * at a point when all managed references are still valid. 
         * Since this class does not directly refer to any system resources, and
         * managed resources can only be cleaned up when the refs are valid, we
         * only clean-up if explicitly called.
         * 
         * Proper usage of this class is to call Close() as soon as the port isn't 
         * needed anymore (as in MainForm.cs). However, if Dispose isn't called here, 
         * then the automatic Dispose call from finalization of the SerialPort 
         * itself will free the OS handle to the serial port.
         * 
         * Note that we would also want to call GCSuppressFinalize (regardless of the
         * value of disposing) if this class were finalizable (i.e. if it had a finalizer).
         * In this case, there are no direct refs to an unmanaged resource, so custom finalization
         * code isn't necessary.
         */
        private void Dispose(bool disposing)
        {
            if (!disposed && disposing && com != null && com.IsOpen)
            {
                Reset();
                com.Close();
                CloseThreads();

                // Keep us from calling resetting or closing multiple times
                disposed = true;
            }
        }
    }

}